module Tabuleiro where

import Graphics.Gloss

import Rotacao
import Tetromino
import Matriz
import Vetor
import Graficos

-- =========================================================================
-- Esse arquivo define a base do tetris, isso eh os tipos de dados e as funcoes
-- referentes a esses dados.
-- =========================================================================

-- =========================================================================
-- IMPORTANTE: Apesar de nao ser o usual na matematica, o Tabuleiro do
-- do tetris sera programado pensando na origem como sendo o seu canto
-- superior esquerdo, sendo que o eixo X cresce para direita e o eixo Y
-- cresce para BAIXO. Esse eh o padrao nas outras engines de programacao
-- de jogos 2d, portanto, sera adotado aqui tambem
-- =========================================================================

-- =========================================================================
-- O tabuleiro eh a parte do jogo com a qual o jogador interage, i.e. onde os tetrominos ficam
data Tabuleiro = Tabuleiro { _grid      :: Matriz (Maybe Bloco) -- Estado atual do tabuleiro, com os tetrominos que ja cairam
                           , _tetromino :: Maybe Tetromino      -- Tetromino atual que o jogador tem controle
                           , _ghost     :: Maybe Tetromino      -- Esse tetromino fica na posicao que o tetromino acima ficaria
                           }                                    -- caso o jogador fizesse um hard drop

-- =========================================================================
-- Como mostrar um tabuleiro no console: Mostre seu grid (obs o tetromino nao
-- eh colocada no grid)
instance Show Tabuleiro where
    show (Tabuleiro g _ _) = show g

instance Eq Tabuleiro where
   (Tabuleiro g1 Nothing _) == (Tabuleiro g2 Nothing _)         = g1 == g2
   (Tabuleiro _ Nothing _) == (Tabuleiro _ (Just x) _)          = False
   (Tabuleiro _ (Just x) _) == (Tabuleiro _ Nothing _)          = False
   (Tabuleiro g1 (Just tet1) _) == (Tabuleiro g2 (Just tet2) _) =
      g1 == g2 && tet1 == tet2 && (_posicao tet1) == (_posicao tet2)

-- =========================================================================
-- Gera o ghost do tetromino atual
setTetromino :: Tabuleiro -> Tetromino -> Tabuleiro
setTetromino tabuleiro tetro = do
   let ghostNovo     = geraGhost (_grid tabuleiro) (Just tetro) (Vetor2D 0 0)
   let tabuleiroNovo = Tabuleiro (_grid tabuleiro) (Just tetro) ghostNovo

   if   tabuleiroEhValido tabuleiroNovo
   then tabuleiroNovo
   else tabuleiro

-- =========================================================================
-- Gera o ghost do tetromino atual
geraGhost :: Matriz (Maybe Bloco) -> Maybe Tetromino -> Vetor2D -> Maybe Tetromino
geraGhost _ Nothing _ = Nothing
geraGhost grid (Just tetro) offset
   | offsetValido = geraGhost grid (Just tetro) (offset + vetor1y)
   | otherwise    = Just tetro'
      where
         tetro'       = Tetromino (_layout tetro) (posicao' - vetor1y) (_orientacao tetro)
         offsetValido = tabuleiroEhValido t'
         t'           = Tabuleiro grid (Just tet') Nothing
         tet'         = Tetromino (_layout tetro) (posicao') (_orientacao tetro)
         posicao'     = (_posicao tetro) + offset
         vetor1y      = Vetor2D 0 1

-- =========================================================================
-- Operacao de integrar um tetromino no tabuleiro, i.e., fixar o tetromino no lugar
-- caso ela nao consiga mais cair e tirar o controle do jogador daquele tetromino
-- a funcao nao verifica se o tetromino esta ocupando o mesmo espaco que outro 
-- bloco, pois outras funcoes vao garantir que isso nunca aconteca
integraTetromino :: Tabuleiro -> Tabuleiro
integraTetromino (Tabuleiro g Nothing Nothing) = Tabuleiro g Nothing Nothing
integraTetromino t@(Tabuleiro g (Just p) _)
    | podeCair  = t                                      -- o tetromino ainda pode cair mais, nao integre ele no tabuleiro
    | otherwise = Tabuleiro novoGrid (Nothing) (Nothing) -- o tetromino nao pode mais cair, pode integrar ele no tabuleiro
    where
         novoGrid      = integraLayoutTetromino (getMatriz $ _layout p) (_y $ _posicao p) (_x $ _posicao p) g
         (_, podeCair) = moveTetroBaixo t

-- Coloca o tetromino no grid do tabuleiro
integraLayoutTetromino :: [[Maybe Bloco]] -> Int -> Int -> Matriz (Maybe Bloco) -> Matriz (Maybe Bloco)
integraLayoutTetromino [] _ _ m     = m
integraLayoutTetromino (l:ls) i j m = integraLinhaTetromino l i j $ integraLayoutTetromino ls (i+1) j m

-- Integra uma das linhas do layout do tetromino no grid do tabuleiro
integraLinhaTetromino :: [Maybe Bloco] -> Int -> Int -> Matriz (Maybe Bloco) -> Matriz (Maybe Bloco)
integraLinhaTetromino []     _ _ m = m
integraLinhaTetromino (b:bs) i j m = integraBlocoTetromino b i j $ integraLinhaTetromino bs i (j+1) m

-- Coloca um bloco do layout do tetromino no grid do tabuleiro
integraBlocoTetromino :: Maybe Bloco -> Int -> Int -> Matriz (Maybe Bloco) -> Matriz (Maybe Bloco)
integraBlocoTetromino Nothing _ _ m = m
integraBlocoTetromino b i j m       = colocaEm b i j m

-- =========================================================================
-- Verifica se todos os blocos de um tetromino estao dentro do tabuleiro
tetrominoEstaNoTabuleiro :: Tabuleiro -> Bool
tetrominoEstaNoTabuleiro (Tabuleiro _ Nothing _)    = True
tetrominoEstaNoTabuleiro (Tabuleiro g (Just tet) _) = do
   let coords = coordsTetromino (_layout tet) (_posicao tet) -- Pega as coordenadas (y, x) dos blocos do tetromino
   xsMaioresZero coords && xsMenoresLargura coords && ysMenoresAltura coords && ysMaioresZero coords
      where
         xsMaioresZero    = all (\(_, x) -> x >= 0)              -- Verifica se todas sao maiores que zero
         xsMenoresLargura = all (\(_, x) -> x < larguraMatriz g) -- Verifica se todas sao menores que a largura
         ysMenoresAltura  = all (\(y, _) -> y < alturaMatriz g)  -- Verifica se todas sao menores que a altura
         ysMaioresZero    = all (\(y, _) -> y >= -1)

-- =========================================================================
-- Verifica se todos os blocos de um tetromino nao colidem com nenhum bloco
-- do tabuleiro
tetrominoNaoColide :: Tabuleiro -> Bool
tetrominoNaoColide (Tabuleiro _ Nothing _)       = True
tetrominoNaoColide (Tabuleiro g (Just tet) _)  = do
   let coords = coordsTetromino (_layout tet) (_posicao tet) -- Pega as coordenadas (y, x) dos blocos do tetromino
   all (avaliaTuplas) coords                                 -- Verifica se nenhuma colide com um bloco do tabuleiro
      where
         avaliaTuplas (-1, j) = True
         avaliaTuplas (i, j) = valorEm g i j == Nothing 

-- =========================================================================
-- Verifica se um determinado offset na posicao do tetromino eh valido, i.e.
-- nao faz com que ele saia do tabuleiro nem colida com blocos que ja estao
-- no tabuleiro
tabuleiroEhValido :: Tabuleiro -> Bool
tabuleiroEhValido (Tabuleiro _ Nothing _) = True
tabuleiroEhValido t                       = tetrominoEstaNoTabuleiro t && tetrominoNaoColide t

-- =========================================================================
-- Aplica um offset recebido na peca atual, caso seja possivel. Sempre devolve
-- um tabuleiro com o ghost do tetromino atual.
aplicaOffset :: Tabuleiro -> Vetor2D -> (Tabuleiro, Bool)
aplicaOffset t@(Tabuleiro _ Nothing _) _       = (t, True)
aplicaOffset (Tabuleiro g (Just tet) _) offset = 
   if   tabuleiroEhValido novoTabuleiro         -- Verifica se o novo tabuleiro eh valido
   then (novoTabuleiro, True)                   -- Se sim, retorna ele junto com o True, indicando um movimento bem sucedido
   else ((Tabuleiro g (Just tet) ghost), False) -- Se nao, retorna o tabuleiro velho com um false, indicando um movimento mal sucedido
   where
      novoTabuleiro = Tabuleiro g (Just tet') ghost'                       -- Gera o novo tabuleiro com o tetromino deslocado
      tet'          = Tetromino (_layout tet) (posicao') (_orientacao tet) -- Gera o novo tetromino com a nova posicao
      posicao'      = (_posicao tet) + offset                              -- Gera um vetor novo para a posicao deslocada para a esquerda
      vetorAltura   = Vetor2D 0 0                                          -- A posicao padrao do ghost eh a msm do tetromino
      ghost'        = geraGhost g (Just tet') vetorAltura
      ghost         = geraGhost g (Just tet) vetorAltura

-- =========================================================================
-- Move o tetromino para a esquerda, se conseguir. O booleano retornado indica
-- se o movimento foi bem sucedido, isso sera usado pelo jogo
moveTetroEsq :: Tabuleiro -> (Tabuleiro, Bool)
moveTetroEsq t = aplicaOffset t (Vetor2D (-1) 0)

-- =========================================================================
-- Move o tetromino para a direita, se conseguir
moveTetroDir :: Tabuleiro -> (Tabuleiro, Bool)
moveTetroDir t = aplicaOffset t (Vetor2D 1 0)

-- =========================================================================
-- Faz com que o tetromino atual caia um bloco, se conseguir
moveTetroBaixo :: Tabuleiro -> (Tabuleiro, Bool)
moveTetroBaixo t = aplicaOffset t (Vetor2D 0 1)

-- =========================================================================
-- Aplica o hard drop, i.e., coloca o tetromino na posica y mais embaixo
-- possivel e integra no tabuleiro
hardDrop :: Tabuleiro -> Tabuleiro
hardDrop t@(Tabuleiro _ Nothing _)  = t
hardDrop (Tabuleiro grid _ ghost) = do
   Tabuleiro grid ghost ghost -- Substitui o tetromino atual pelo ghost

-- =========================================================================
-- Tenta aplicar a funcao de rotacao no tetromino, fazendo o wallkick apropriado
-- quando for necessario. Caso nem a rotacao simples nem o wallkick sejam sucedidos
-- nao faz nada.
aplicaRotacao :: Tabuleiro -> (Tetromino -> Tetromino) -> (Tabuleiro, Bool)
aplicaRotacao t@(Tabuleiro _ Nothing _) _                = (t, True)
aplicaRotacao t@(Tabuleiro g (Just tet) _) funcaoRotacao = do
   let funcaoTestes = escolheTestes tet -- Pega a funcao de testes apropriada para o tetromino atual

   let tetroRodado = funcaoRotacao tet       -- Rotaciona o tetromino, aplicando a rotacao passada
   let oInicial    = _orientacao tet         -- Orientacao antes da rotacao do tetromino
   let oFinal      = _orientacao tetroRodado -- Orientacao depois da rotacao do tetromino

   let offsets      = funcaoTestes oInicial oFinal             -- Offsets que devem ser testados, de acordo com as orientacoes e o tetromino atual
   let novoTabBase  = Tabuleiro g (Just tetroRodado) (Nothing) -- Tabuleiro com o tetromino rodado, o ghst de cada tabuleiro sera gerado abaixo
   let tabsOffset   = map (aplicaOffset novoTabBase) offsets   -- Aplicando todos os offsets e gerando os ghost correspondentes

   if   any (\(_, bool) -> bool) tabsOffset
   then head $ filter (\(_, bool) -> bool) tabsOffset -- Se tiver um ou mais tabuleiros validos, devolva o primeiro deles
   else (t, False)                                    -- Caso contrario, devolva o tabuleiro atual

-- Rotaciona o tetromino para a esquerda (sentido anti-horario). Caso a rotacao 
-- falhe o jogo tenta fazer um Wall Kick, i.e. transladar a peca para as casas 
-- vizinhas para tentar encaixa-la. Caso nenhum das duas coisas seja possivel
-- nada acontece.
-- O primeiro parametro do retorno eh o tabuleiro atualizado.
-- O segundo parametro do retorno diz se a rotacao foi bem sucedida
-- O terceiro Parametro do retorno diz se houve ou nao um T-spin
aplicaRotacaoEsq :: Tabuleiro -> (Tabuleiro, Bool)
aplicaRotacaoEsq t = aplicaRotacao t rotTetroEsq
   
-- =========================================================================
-- Rotaciona o tetromino para a direita (sentido horario). Caso a rotacao falhe
-- o jogo tenta fazer um Wall Kick, i.e. transladar a peca para as casas 
-- vizinhas para tentar encaixa-la. Caso nenhum das duas coisas seja possivel
-- nada acontece.
-- O primeiro parametro do retorno eh o tabuleiro atualizado.
-- O segundo parametro do retorno diz se a rotacao foi bem sucedida
-- O terceiro Parametro do retorno diz se houve ou nao um T-spin
aplicaRotacaoDir :: Tabuleiro -> (Tabuleiro, Bool)
aplicaRotacaoDir t = aplicaRotacao t rotTetroDir

-- =========================================================================
-- Verifica se uma ou mais linhas estao cheias e limpa essas linhas, retornando
-- a quantidade de linhas limpadas.
limpaLinhas :: Tabuleiro -> (Tabuleiro, Int)
limpaLinhas t@(Tabuleiro _ (Just _) _ ) = (t, 0)
limpaLinhas t@(Tabuleiro grid Nothing _) = do
   let alturaInicial = alturaMatriz grid -- Salva a altura inicial

   let gridSemLinhasCheias   = tiraLinhasCheias grid            -- Limpa as linhas cheias
   let alturaSemLinhasCheias = alturaMatriz gridSemLinhasCheias -- Salva a altura nova

   let diffAlturas = alturaInicial - alturaSemLinhasCheias          -- Quantidade de linhas que foram tiradas
   let gridFinal   = completaMatriz gridSemLinhasCheias diffAlturas -- Coloca linhas vazias em cima da matriz, para compensar as linhas removidas

   let tabuleiroNovo = Tabuleiro gridFinal Nothing Nothing -- Tabuleiro novo, com a grid limpa e a altura ajustada

   if   diffAlturas == 0             
   then (t, 0)                       -- Caso nenhuma linha tenha sido removida, retorna o tabuleiro atual
   else (tabuleiroNovo, diffAlturas) -- Caso contrario, retorna o novo tabuleiro
   

-- =========================================================================
-- Remove todas as linhas que nao contenham nenhum nothing da matriz
tiraLinhasCheias :: Matriz (Maybe Bloco) -> Matriz (Maybe Bloco)
tiraLinhasCheias (Matriz [])             = Matriz []
tiraLinhasCheias (Matriz (linha:linhas)) = 
   if   all (/= Nothing) linha
   then tiraLinhasCheias $ Matriz linhas                       -- Caso a linha atual esteja cheia, remove a linha atual
   else (Matriz [linha]) <> (tiraLinhasCheias $ Matriz linhas) -- caso contrario, nao faca nada

-- =========================================================================
-- Remove todas as linhas que nao contenham nenhum nothing da matriz
completaMatriz :: Matriz (Maybe Bloco) -> Int -> Matriz (Maybe Bloco)
completaMatriz (Matriz m) 0 = Matriz m                 -- Se precisar adicionar 0 linhas, nao faca nada
completaMatriz (Matriz m) n = completaMatriz mat (n-1) -- Caso contrario, adicione uma linha
   where
      mat        = (Matriz [linhaNova]) <> (Matriz m)
      linhaNova  = take largura $ repeat Nothing
      largura    = larguraMatriz $ Matriz m

-- =========================================================================
-- Desenha o tabuleiro, com o tetromino atual e o ghost
desenhaTabuleiro :: Float -> Tabuleiro -> Picture
desenhaTabuleiro tamanhoBloco (Tabuleiro (Matriz (grid)) tet ghost)= do
   -- Parametros para o background
   let largura = fromIntegral $ larguraMatriz $ Matriz grid
   let altura  = fromIntegral $ alturaMatriz $ Matriz grid

   -- Essa eh a cor com a qual as divisoes do tabuleiro vao aparecer
   let background = montaRetangulo 0 0 (largura*tamanhoBloco-1) (altura*tamanhoBloco-1) 2 white white 

   -- Desenhando o tabuleiro
   let tabuleiroBG  = geraBG $ Matriz grid
   let tabBgGrafico = desenhaLayout tabuleiroBG 0 0 tamanhoBloco False False
   let tabuleiroGrafico = desenhaLayout (Matriz grid) 0 0 tamanhoBloco False True

   -- Desenhando o tetromino atual
   let tetroGrafico = if tet == Nothing then Blank else desenhaLayout (_layout $ fromJust tet) xT yT tamanhoBloco False False

   -- Desenhando o tetromino ghost 
   let ghostGrafico = if ghost == Nothing then Blank else desenhaLayout (_layout $ fromJust ghost) xG yG tamanhoBloco True False

   Pictures [background, tabBgGrafico, tabuleiroGrafico, ghostGrafico, tetroGrafico]
      where
         fromJust (Just x) = x
         xT = (fromIntegral $ _x $ _posicao $ fromJust tet)*tamanhoBloco
         yT = (fromIntegral $ _y $ _posicao $ fromJust tet)*(-tamanhoBloco)
         xG = (fromIntegral $ _x $ _posicao $ fromJust ghost)*tamanhoBloco
         yG = (fromIntegral $ _y $ _posicao $ fromJust ghost)*(-tamanhoBloco)

-- =========================================================================
-- Gera um tabuleiro em que todos os espacos estao preenchidos por blocos
-- pretos. Ele sera o background do tabuleiro real. Caso ele nao existisse
-- nos lugares onde nao temos nada teriamos a cor solida do background
geraBG :: Matriz (Maybe Bloco) -> Matriz (Maybe Bloco)
geraBG (Matriz [])             = Matriz []
geraBG (Matriz (linha:linhas)) = linhaNova <> linhasNovas
   where
      linhaNova   = Matriz ([take largura $ repeat (Just $ Bloco black)])
      linhasNovas = geraBG $ Matriz linhas
      largura     = larguraMatriz $ Matriz [linha]