module Tetromino where

import Graphics.Gloss

import Matriz
import Vetor
import Graficos

-- =========================================================================
-- Esse arquivo contem as definicoes relativas aos Tetrominoes, que sao as
-- pecas do tetris.
-- =========================================================================

-- =========================================================================
-- O Bloco eh a unidade que forma cado tetromino
data Bloco = Bloco { _cor :: Color}

-- =========================================================================
-- A orientacao diz como o Bloco esta rodado, sendo Zero a orientacao
-- inicial de cada Tetromino, Left/Right as orientacoes apos realizar
-- uma rotacao para a esquerda/direita e dois a orientacao apos realizar
-- duas da mesma rotacao
data Orientacao = OLeft | OZero | ORight | ODois deriving (Enum, Eq)

-- =========================================================================
-- Tetrominoes sao as tetrominos do tetris
data Tetromino  = Tetromino { _layout     :: Matriz (Maybe Bloco) -- Formato do tetromino, tambem eh chamado de caixa do tetromino
                            , _posicao    :: Vetor2D              -- Posicao do tetromino no tabuleiro
                            , _orientacao :: Orientacao           -- Quqis rotacoes foram aplicadas na tetromino
                            }
-- Note que a posicao do tetromino eh a posicao (x,y) da matriz do tabuleiro em que
-- a matriz do tetromino comeca a ser desenhada.

-- =========================================================================
-- Como comparar dois tetrominoes: Compare apenas o layout deles, sem se
-- importar com a rotacao
instance Eq Tetromino where
    t1 == t2
        | _layout t1 == _layout t2                = True
        | _layout t1 == _layout (rotTetroDir t2)  = True
        | _layout t1 == _layout (rotTetroEsq t2)  = True
        | _layout t1 == _layout (rotTetroDois t2) = True
        | otherwise                               = False

-- Como mostrar um tetromino
instance Show Tetromino where
    show t = show $ _layout t

-- =========================================================================
-- Como comparar dois blocos: Dois blocos sao sempre iguais.
-- Apesar da cor de um bloco poder ser diferente da outra, o programa nao
-- precisa enxergar nenhuma diferenca entre eles, apenas o usuario
instance Eq Bloco where
    _ == _ = True

-- =========================================================================
-- Como mostrar um bloco no terminal
instance Show Bloco where
    show _ = "Bloco"

-- =========================================================================
-- Rotaciona uma tetromino para a direita
rotTetroDir :: Tetromino -> Tetromino
rotTetroDir tet@(Tetromino l p o) 
    | tet == tetroO = tet
    | o == ODois    = Tetromino (rotacionaDir l) p OLeft
    | otherwise     = Tetromino (rotacionaDir l) p (succ o)

-- =========================================================================
-- Rotaciona uma tetromino para a esquerda
rotTetroEsq :: Tetromino -> Tetromino
rotTetroEsq tet@(Tetromino l p o) 
    | tet == tetroO = tet
    | o == OLeft    = Tetromino (rotacionaEsq l) p ODois
    | otherwise     = Tetromino (rotacionaEsq l) p (pred o)

-- =========================================================================
-- Rotaciona uma tetromino duas vezes
rotTetroDois :: Tetromino -> Tetromino
rotTetroDois tet@(Tetromino l p o) 
    | tet == tetroO = tet
    | o == OLeft    = Tetromino ((rotacionaDir . rotacionaDir) l) p ORight
    | o == ORight   = Tetromino ((rotacionaDir . rotacionaDir) l) p OLeft
    | o == OZero    = Tetromino ((rotacionaDir . rotacionaDir) l) p ODois
    | o == ODois    = Tetromino ((rotacionaDir . rotacionaDir) l) p OZero

-- =========================================================================
-- Gera uma lista com pares (y, x) que correspondem a todas as posicoes
-- nao vazias do layout de um tetromino. O parametro Vetor2D eh o deslocamento
-- dele, assim podemos usar essa funcao para encontrar as coordenadas absolutas
-- de tetromino que nao estao vazias
coordsTetromino :: Matriz (Maybe Bloco) -> Vetor2D -> [(Int, Int)]
coordsTetromino (Matriz []) _               = []
coordsTetromino (Matriz (linha:linhas)) pos = coordsLinha ++ coordsMatriz
        where
         -- Pares ordenados
            coordsLinha  = [(n, j) | (j,_) <- posicoesPreenchidas linha]
            coordsMatriz = coordsTetromino (Matriz linhas) (Vetor2D m (n+1))
         -- Variaveis auxiliares
            n = _y pos
            m = _x pos
         -- Funcoes auxiliares
            posicoesPreenchidas xs = filter filtraTupla $ zip [m..] xs
            filtraTupla (_, mb) = mb /= Nothing

-- =========================================================================
-- Desenha o layout do tetromino
desenhaLayout :: Matriz (Maybe Bloco) -> Float -> Float -> Float -> Bool -> Bool -> Picture
desenhaLayout (Matriz []) _ _ _ _ _                                         = Blank
desenhaLayout (Matriz (linha:linhas)) x y tamanhoLado ehGhost estaIntegrado = Pictures [linhaAtual, proxLinhas]
    where
        linhaAtual = desenhaLinha linha x y tamanhoLado ehGhost estaIntegrado
        proxLinhas = desenhaLayout (Matriz linhas) x y' tamanhoLado ehGhost estaIntegrado
        y'         = y - tamanhoLado

-- =========================================================================
-- Desenha uma linha do tetromino
desenhaLinha :: [Maybe Bloco] -> Float -> Float -> Float -> Bool -> Bool -> Picture
desenhaLinha [] _ _ _ _ _                                 = Blank
desenhaLinha (b:bs) x y tamanhoLado ehGhost estaIntegrado = Pictures [bloco, blocos]
    where
        bloco  = desenhaBloco b x y tamanhoLado ehGhost estaIntegrado
        blocos = desenhaLinha bs x' y tamanhoLado ehGhost estaIntegrado
        x'     = x+tamanhoLado

-- =========================================================================
-- Desenha um bloco
desenhaBloco :: Maybe Bloco -> Float -> Float -> Float -> Bool -> Bool -> Picture
desenhaBloco Nothing _ _ _ _ _                                        = Blank
desenhaBloco (Just (Bloco cor)) x y tamanhoLado ehGhost estaIntegrado = montaQuadrado (x) (y) (tamanhoLado-1) 0 cor' cor'
    where
        cor' = if ehGhost then light $ withAlpha 0.15 cor
          else if estaIntegrado then dark cor else cor

-- =========================================================================
-- Desenha o hold
desenhaHold :: Matriz (Maybe Bloco) -> Float -> Picture
desenhaHold tetro tamanhoBloco = do
    -- Desenhando o bloco hold
    let holdGrafico = desenhaLayout tetro tamanhoBloco (-tamanhoBloco) tamanhoBloco False False

    -- Desenha o background
    let background = montaRetangulo 0 0 (6*tamanhoBloco) (4*tamanhoBloco) 2 white black

    -- Desenhando o titulo e seu background
    let tituloBg      = montaRetangulo 0 0 (6*tamanhoBloco) (-(1.5*tamanhoBloco)) 2 white white
    let tituloGrafico = Translate 0 3 $ Scale (tamanhoBloco/100) (tamanhoBloco/100) $ text "Hold (W)"

    Pictures [background, tituloBg, tituloGrafico, holdGrafico]

-- =========================================================================
-- =========================================================================
-- Constantes dos tetrominoes

-- ================================================================
-- Posicao inicial dos blocos, a posicao de todos eh igual, com excessao
-- do bloco O, que comeca um pouco mais a direita por ser mais estreito
-- que os outros
posInicial :: Vetor2D
posInicial  = Vetor2D 3 (-1)

-- ================================================================
-- Os blocos que compoe cada um dos Tetrominoes, cada um com sua respectiva
-- cor
blocoI, blocoJ, blocoL, blocoO, blocoS, blocoT, blocoZ :: Bloco
blocoI = Bloco cyan
blocoJ = Bloco blue
blocoL = Bloco orange
blocoO = Bloco yellow
blocoS = Bloco green
blocoT = Bloco magenta
blocoZ = Bloco red

-- ================================================================
-- O layout de cada tetromino eh o formato do Tetromino, que sera colocado no
-- tabuleiro
layoutI, layoutJ, layoutL, layoutO, layoutS, layoutT, layoutZ, layoutEmpty :: [[Maybe Bloco]]
layoutI = [[Nothing,     Nothing,     Nothing,     Nothing    ],
           [Just blocoI, Just blocoI, Just blocoI, Just blocoI],
           [Nothing,     Nothing,     Nothing,     Nothing    ],
           [Nothing,     Nothing,     Nothing,     Nothing    ]]

layoutJ = [[Just blocoJ, Nothing,     Nothing    ],
           [Just blocoJ, Just blocoJ, Just blocoJ],
           [Nothing,     Nothing,     Nothing    ]]

layoutL = [[Nothing,     Nothing,     Just blocoL],
           [Just blocoL, Just blocoL, Just blocoL],
           [Nothing,     Nothing,     Nothing    ]]

layoutO = [[Nothing, Just blocoO, Just blocoO, Nothing],
           [Nothing, Just blocoO, Just blocoO, Nothing],
           [Nothing, Nothing,     Nothing,     Nothing]]

layoutS = [[Nothing,     Just blocoS, Just blocoS],
           [Just blocoS, Just blocoS, Nothing    ],
           [Nothing,     Nothing,     Nothing    ]]

layoutT = [[Nothing,     Just blocoT, Nothing    ],
           [Just blocoT, Just blocoT, Just blocoT],
           [Nothing,     Nothing,     Nothing    ]]

layoutZ = [[Just blocoZ, Just blocoZ, Nothing    ],
           [Nothing,     Just blocoZ, Just blocoZ],
           [Nothing,     Nothing,     Nothing    ]]

layoutEmpty = [[Nothing, Nothing, Nothing],
               [Nothing, Nothing, Nothing],
               [Nothing, Nothing, Nothing]]

-- ================================================================
-- Essa eh a definicao do Tetromino, contendo sua rotacao, seu layout
-- e sua posicao inicial
tetroI, tetroJ, tetroL, tetroO, tetroS, tetroT, tetroZ :: Tetromino
tetroI = Tetromino (Matriz layoutI) posInicial OZero
tetroJ = Tetromino (Matriz layoutJ) posInicial OZero
tetroL = Tetromino (Matriz layoutL) posInicial OZero
tetroO = Tetromino (Matriz layoutO) posInicial OZero
tetroS = Tetromino (Matriz layoutS) posInicial OZero
tetroT = Tetromino (Matriz layoutT) posInicial OZero
tetroZ = Tetromino (Matriz layoutZ) posInicial OZero

-- ================================================================
-- Lista com todos os tetrominoes
todosTetrominoes :: [Tetromino]
todosTetrominoes = [tetroI, tetroJ, tetroL, tetroO, tetroS, tetroT, tetroZ]